import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.awt.image.BufferedImage;

class NeighborGrid extends JPanel implements MouseListener
{
  private static int gridWidth = 200;
  private static int gridHeight = 200;
  private static int gridColumns = 5;
  private static int gridRows = 3;
  private static int zoomLevel = 25;

  TriGridSpinner triGridSpinner[][];
  private BufferedImage image;
  private int minimumWeight;
  private int maximumWeight;

  public NeighborGrid(int minimumWeight, int maximumWeight)
  {
    this.minimumWeight = minimumWeight;
    this.maximumWeight = maximumWeight;
    triGridSpinner = new TriGridSpinner[gridColumns][gridRows];
    int spinnerWidth = 2 * zoomLevel;
    int spinnerHeight = 2 * zoomLevel;
    int xOffset = (gridWidth - (gridColumns * spinnerWidth + spinnerWidth) / 2) / 2;
    int yOffset = (gridHeight - gridRows * spinnerHeight) / 2;
    int x, y;
    boolean pointingUp = true;
    for (int i = 0; i < gridColumns; i++)
    {
      x = xOffset + i * spinnerWidth / 2;
      for (int j = 0; j < gridRows; j++)
      {
        y = yOffset + j * spinnerHeight;
        if ((j == 2) && ((i == 0) || (i == 4)))
          triGridSpinner[i][j] = null;
        else
          triGridSpinner[i][j] = new TriGridSpinner(x, y, spinnerWidth, spinnerHeight, pointingUp, 0);
        pointingUp = !pointingUp;
      }
    }
    addMouseListener(this);
    setPreferredSize(new Dimension(gridWidth, gridHeight));
  }

  public void paintComponent(Graphics g)
  {
    super.paintComponent(g);
    if (image == null)
      image = new BufferedImage(getWidth(), getHeight(), BufferedImage.TYPE_INT_RGB);
    drawGrid();
    g.drawImage(image, 0, 0, null);
  }

  private void drawGrid()
  {
    Graphics g = image.getGraphics();
    g.setFont(new Font("Monospaced", Font.PLAIN, 14));
    g.setColor(getBackground());
    g.fillRect(0, 0, image.getWidth(), image.getHeight());
    boolean pointingUp = true;
    for (int i = 0; i < gridColumns; i++)
      for (int j = 0; j < gridRows; j++)
        if ((j == 2) && ((i == 0) || (i == 4)))
          continue;
        else
          triGridSpinner[i][j].draw(g);
    g.dispose();
  }

  public int getValue(int r, int c)
  {
    if (triGridSpinner[c][r] != null)
      return triGridSpinner[c][r].getWeight();
    else
      return 0;
  }

  public void setValue(int r, int c, int value)
  {
    if (triGridSpinner[c][r] != null)
      triGridSpinner[c][r].setWeight(value);
  }

  public void mousePressed(MouseEvent e)
  {
    for (int i = 0; i < gridColumns; i++)
      for (int j = 0; j < gridRows; j++)
      {
        if (triGridSpinner[i][j] == null)
          continue;
        if (triGridSpinner[i][j].contains(e.getX(), e.getY()))
        {
          int weight = triGridSpinner[i][j].getWeight();
          if (e.isShiftDown())
            if (weight == minimumWeight)
              weight = maximumWeight;
            else
              weight--;
          else
            if (weight == maximumWeight)
              weight = minimumWeight;
            else
              weight++;
          triGridSpinner[i][j].setWeight(weight);
          repaint();
          return;
        }
      }
  }

  public void mouseExited(MouseEvent e) {}
  public void mouseClicked(MouseEvent e) {}
  public void mouseEntered(MouseEvent e) {}
  public void mouseReleased(MouseEvent e) {}
}