import java.awt.*;
import javax.swing.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;

public class TriangularCell extends JPanel implements ActionListener, MouseListener, MouseMotionListener, KeyListener
{
  public static void main(String[] args)
  {
    mainFrame = new JFrame("Triangular Cell");
    mainFrame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    mainFrame.add(new TriangularCell());
    Toolkit k = Toolkit.getDefaultToolkit();
    Dimension s = k.getScreenSize();
    mainFrame.setLocation((s.width - width) / 2, (s.height - height) / 2);
    mainFrame.setSize(width, height);
    mainFrame.setVisible(true);
  }

  private static int width = 1000;
  private static int height = 800;
  private static int columns = 1200;
  private static int rows = 1000;

  public static JFrame mainFrame;

  private int cells[][];
  private int cellsCopy[][];
  private TriGrid triGrid;
  private ColorSelector colorSelector;
  private StatusBar statusBar;
  private Selector selector;
  private NewWorld newWorld;
  private Analyzer analyzer;
  private JToolBar toolBar;
  private JButton setNewWorldButton;
  private JButton openFileButton;
  private JButton saveFileButton;
  private JButton newRuleButton;
  private JButton runStopButton;
  private JButton stepButton;
  private JButton clearButton;
  private JButton randomButton;
  private JButton randomOptionsButton;
  private JButton analyzerButton;
  private JButton rulesButton;
  private JButton ruleOptionsButton;
  private JButton changeColorsButton;
  private JButton clearInsideButton;
  private JButton clearOutsideButton;
  private JButton flipXButton;
  private JButton flipYButton;
  private JButton rotateClockwiseButton;
  private JButton rotateAntiClockwiseButton;
  private JButton selectAllButton;
  private JButton centerButton;
  private JButton saveButton;
  private JButton selectionRandomButton;
  private JButton moveCopyButton;
  private Thread runThread;
  private boolean running = false;
  private int delay = 0;
  private boolean analyzing = false;
  private int cycle = 0;
  private File currentDirectory;
  private Rule rule;
  private int randomDensity = 30;
  private boolean randomClear = true;
  private boolean randomAllStates = true;
  private Patterns patterns;

  public TriangularCell()
  {
    UIManager.LookAndFeelInfo[] infos = UIManager.getInstalledLookAndFeels();
    try
    {
      UIManager.setLookAndFeel(infos[0].getClassName());
      SwingUtilities.updateComponentTreeUI(this);
    }
    catch (Exception e)
    {
      e.printStackTrace();
    }
    cells = new int[columns][rows];
    cellsCopy = new int[columns][rows];
    for (int c = 0; c < columns; c++)
      for (int r = 0; r < rows; r++)
        cells[c][r] = 0;
    statusBar = new StatusBar(columns, rows);
    triGrid = new TriGrid(cells, columns, rows);
    colorSelector = new ColorSelector(mainFrame, this, triGrid);
    selector = new Selector(this, triGrid, cells, columns, rows);
    triGrid.setComponents(this, selector, statusBar, colorSelector);
    setLayout(new BorderLayout(3, 3));
    makeToolBar();
    add(toolBar, BorderLayout.NORTH);
    add(triGrid, BorderLayout.CENTER);
    add(statusBar, BorderLayout.SOUTH);
    triGrid.addMouseListener(this);
    triGrid.addMouseMotionListener(this);
    addKeyListener(this);
    setFocusable(true);
    rule = new RuleTableRule(mainFrame, this);
  }

  private void makeToolBar()
  {
    toolBar = new JToolBar();
    setNewWorldButton = getNewButton("NewWorld", "set cellular automata parameters");
    toolBar.addSeparator();
    openFileButton = getNewButton("OpenFile", "read pattern from file to selection");
    saveFileButton = getNewButton("SaveFile", "write pattern to file");
    toolBar.addSeparator();
    newRuleButton = getNewButton("NewRule", "test new random rule");
    toolBar.addSeparator();
    runStopButton = getNewButton("Run", "run the simulation");
    stepButton = getNewButton("Step", "step the simulation");
    toolBar.addSeparator();
    clearButton = getNewButton("Clear", "clear world");
    toolBar.addSeparator();
    randomButton = getNewButton("Random", "set world to random states");
    randomOptionsButton = getNewButton("RandomOptions", "change random options");
    toolBar.addSeparator();
    analyzerButton = getNewButton("Analyzer", "show ship, oscillator or still life characteristics");
    toolBar.addSeparator();
    rulesButton = getNewButton("Rules", "change rule");
    ruleOptionsButton = getNewButton("RuleOptions", "change rule options");
    toolBar.addSeparator();
    changeColorsButton = getNewButton("ChangeColors", "change state colors");
    toolBar.addSeparator();
    clearInsideButton = getNewButton("ClearInside", "clear inside selection");
    clearOutsideButton = getNewButton("ClearOutside", "clear outside selection");
    flipXButton = getNewButton("FlipX", "flip selection about X axis");
    flipYButton = getNewButton("FlipY", "flip selection about Y axis");
    rotateClockwiseButton = getNewButton("Rotate+60", "rotate selection clockwise");
    rotateAntiClockwiseButton = getNewButton("Rotate-60", "rotate selection anti clockwise");
    selectAllButton = getNewButton("SelectAll", "select all live cells");
    centerButton = getNewButton("Center", "center selection");
    saveButton = getNewButton("SaveSelection", "write selection to file");
    selectionRandomButton = getNewButton("RandomSelection", "set selection to random states");
    moveCopyButton = getNewButton("Cut", "set selector to copy mode");
  }

  private JButton getNewButton(String name, String hint)
  {
    JButton b = new JButton(new ImageIcon(TriangularCell.class.getResource("Resources/" + name + ".png")));
    b.setBackground(new Color(192, 192, 192));
    b.setToolTipText(hint);
    b.addActionListener(this);
    toolBar.add(b);
    return b;
  }

  private void centerDialog(JDialog dialog)
  {
    Point p = mainFrame.getLocation();
    int x = (int)p.getX() + (mainFrame.getWidth() - dialog.getWidth()) / 2;
    int y = (int)p.getY() + (mainFrame.getHeight() - dialog.getHeight()) / 2;
    dialog.setLocation(x, y);
  }

  public void setNewWorld()
  {
    newWorld = new NewWorld(mainFrame, columns, rows);
    centerDialog(newWorld);
    newWorld.setVisible(true);
    if (newWorld.hasChanged())
    {
      columns = newWorld.getColumns();
      rows = newWorld.getRows();
      width = mainFrame.getWidth();
      height = mainFrame.getHeight();
      makeNewTriangularCell();
    }
  }

  private static void makeNewTriangularCell()
  {
    mainFrame.setVisible(false);
    mainFrame.dispose();
    main(null);
  }

  public ColorSelector getColorSelector()
  {
    return colorSelector;
  }

  public StatusBar getStatusBar()
  {
    return statusBar;
  }

  private void clear()
  {
    if (selector.getState() != Selector.State.None)
      selector.remove();
    selector.clearOutside();
    triGrid.resetOffset();
    statusBar.removePatternName();
    cycle = 0;
    statusBar.setCycle(cycle);
  }

  private boolean isClear()
  {
    for (int c = 0; c < columns; c++)
      for (int r = 0; r < rows; r++)
        if (cells[c][r] > 0)
          return false;
    return true;
  }

  private void analyzePattern()
  {
    analyzer = new Analyzer(mainFrame, this, triGrid, selector.stateMask);
    if (analyzer.save(cells, columns, rows))
      return;
    analyzerButton.setIcon(new ImageIcon(TriangularCell.class.getResource("Resources/Stop.png")));
    analyzerButton.setToolTipText("stop the analyzer");
    cycle = 0;
    runThread = new Thread(analyzer);
    analyzing = true;
    runThread.start();
  }

  public void showAnalysisResults()
  {
    analyzing = false;
    analyzerButton.setIcon(new ImageIcon(TriangularCell.class.getResource("Resources/Analyzer.png")));
    analyzerButton.setToolTipText("show ship, oscillator or still life characteristics");
    centerDialog(analyzer);
    analyzer.setVisible(true);
    triGrid.repaint();
  }

  public boolean isAnalyzing()
  {
    return analyzing;
  }

  private void changeColors()
  {
    centerDialog(colorSelector);
    colorSelector.setVisible(true);
  }
  
  private void setCut(boolean cut)
  {
    if (cut)
    {
      moveCopyButton.setIcon(new ImageIcon(TriangularCell.class.getResource("Resources/Cut.png")));
      moveCopyButton.setToolTipText("set selector to copy mode");
      selector.setCut(true);
    }
    else
    {
      moveCopyButton.setIcon(new ImageIcon(TriangularCell.class.getResource("Resources/Copy.png")));
      moveCopyButton.setToolTipText("set selector to move mode");
      selector.setCut(false);
    }
  }

  private void readPattern()
  {
    JFileChooser fc = new JFileChooser();
    fc.setCurrentDirectory(currentDirectory);
    if (fc.showOpenDialog(this) == JFileChooser.APPROVE_OPTION)
    {
      File f = fc.getSelectedFile();
      currentDirectory = fc.getCurrentDirectory();
      patterns = new Patterns(currentDirectory, f.getName());
      readFile(f);
    }
  }

  private void readFile(File f)
  {
    try
    {
      BufferedReader br = new BufferedReader(new FileReader(f));
      String ruleName = br.readLine().substring(8);
      if (ruleName.equals("Rule Table"))
      {
        rule = new RuleTableRule(mainFrame, this);
        rule.readRule(br);
      }
      else
      {
        br.close();
        return;
      }
      selector.readNewSelection(br);
      br.close();
      statusBar.setPatternName(f.getName());
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }

  private void writePattern()
  {
    if (selector.getState() != Selector.State.GroupSelected)
      return;
    JFileChooser fc = new JFileChooser();
    fc.setCurrentDirectory(currentDirectory);
    if (fc.showSaveDialog(this) == JFileChooser.APPROVE_OPTION)
    {
      try
      {
        File f = fc.getSelectedFile();
        currentDirectory = fc.getCurrentDirectory();
        PrintWriter pw = new PrintWriter(new BufferedWriter(new FileWriter(f)));
        rule.writeRule(pw);
        selector.writeHeader(pw);
        selector.writeSelection(pw);
        pw.close();
        patterns = new Patterns(currentDirectory, f.getName());
        statusBar.setPatternName(f.getName());
      }
      catch (IOException e)
      {
        e.printStackTrace();
      }
    }
  }

  public int getNoStates()
  {
    return rule.getNoStates();
  }

  public int getCell(int c, int r)
  {
    return cellsCopy[(c + columns) % columns][(r + rows) % rows] & selector.stateMask;
  }

  public void step()
  {
    int t[][] = cellsCopy;
    cellsCopy = cells;
    cells = t;
    for (int c = 0; c < columns; c++)
      for (int r = 0; r < rows; r++)
        cells[c][r] = rule.step(c, r);
    selector.setCells(cells);
    triGrid.setCells(cells);
    statusBar.setCycle(++cycle);
  }

  public int[][] analyzerStep()
  {
    step();
    return cells;
  }

  public void changeRandomOptions(JFrame mainFrame)
  {
    RandomOptionsDialog randomOptionsDialog = new RandomOptionsDialog(mainFrame);
    randomOptionsDialog.setData(this);
    centerDialog(randomOptionsDialog);
    randomOptionsDialog.setVisible(true);
    if (randomOptionsDialog.hasChanged())
      randomOptionsDialog.getData(this);
  }

  public int getRandomDensity()
  {
    return randomDensity;
  }

  public boolean getRandomClear()
  {
    return randomClear;
  }

  public boolean getRandomAllStates()
  {
    return randomAllStates;
  }

  public void setRandomDensity(int density)
  {
    randomDensity = density;
  }

  public void setRandomClear(boolean value)
  {
    randomClear = value;
  }

  public void setRandomAllStates(boolean value)
  {
    randomAllStates = value;
  }

  public void random(int columns, int rows, int cells[][])
  {
    Random generator = new Random();
    if (randomClear)
      for (int c = 0; c < columns; c++)
        for (int r = 0; r < rows; r++)
          cells[c][r] = 0;
    for (int c = 0; c < columns; c++)
      for (int r = 0; r < rows; r++)
        if (generator.nextInt(100) < randomDensity)
          if (randomAllStates)
            cells[c][r] = generator.nextInt(rule.getNoStates());
          else
            cells[c][r] = triGrid.getCurrentState();
    triGrid.repaint();
  }

  private void waitForThreadToEnd()
  {
    while (runThread.isAlive())
    {
      try
      {
        Thread.sleep(50);
      }
      catch (InterruptedException ie)
      {
      }
    }
  }

  private void startRunner()
  {
    runStopButton.setIcon(new ImageIcon(TriangularCell.class.getResource("Resources/Stop.png")));
    runStopButton.setToolTipText("stop the simulation");
    cycle = 0;
    statusBar.setCycle(cycle);
    statusBar.setSpeedDelay(delay);
    statusBar.removePopulation();
    Runnable r = new CaRunner(this, triGrid);
    runThread = new Thread(r);
    running = true;
    runThread.start();
  }

  private void stopRunner()
  {
    running = false;
    waitForThreadToEnd();
    statusBar.removeSpeedDelay();
    runStopButton.setIcon(new ImageIcon(TriangularCell.class.getResource("Resources/Run.png")));
    runStopButton.setToolTipText("run the simulation");
    triGrid.repaint();
  }

  public boolean isRunning()
  {
    return running;
  }

  public int getDelay()
  {
    return delay;
  }

  public void actionPerformed(ActionEvent e)
  {
    Object source = e.getSource();
    if (analyzing)
    {
      if (source == analyzerButton)
      {
        analyzing = false;
        waitForThreadToEnd();
        analyzerButton.setIcon(new ImageIcon(TriangularCell.class.getResource("Resources/Analyzer.png")));
        analyzerButton.setToolTipText("show ship, oscillator or still life characteristics");
      }
    }
    else if (running)
    {
      if ((source == runStopButton) || (source == newRuleButton) || (source == randomButton))
        stopRunner();
      if (source == newRuleButton)
      {
        rule.setNewRule();
        random(columns, rows, cells);
        startRunner();
      }
      else if (source == randomButton)
      {
        random(columns, rows, cells);
        startRunner();
      }
    }
    else if (source == runStopButton)
      startRunner();
    else if (source == newRuleButton)
    {
      rule.setNewRule();
      random(columns, rows, cells);
      startRunner();
    }
    else if (source == setNewWorldButton)
      setNewWorld();
    else if (source == openFileButton)
      readPattern();
    else if (source == saveFileButton)
    {
      if ((! isClear()) && (selector.getState() == Selector.State.None))
      {
        selector.selectAll();
        writePattern();
        selector.restoreCells();
        selector.remove();
      }
    }
    else if (source == stepButton)
    {
      step();
      triGrid.repaint();
    }
    else if (source == randomButton)
      random(columns, rows, cells);
    else if (source == randomOptionsButton)
      changeRandomOptions(mainFrame);
    else if (source == clearButton)
      clear();
    else if (source == analyzerButton)
      analyzePattern();
    else if (source == changeColorsButton)
      changeColors();
    else if (source == rulesButton)
      rule.changeRule();
    else if (source == ruleOptionsButton)
      rule.setOptions();
    else if (source == clearInsideButton)
    {
      if (selector.getState() == Selector.State.GroupSelected)
        selector.remove();
    }
    else if (source == clearOutsideButton)
    {
      if (selector.getState() == Selector.State.GroupSelected)
        selector.clearOutside();
    }
    else if (source == flipXButton)
      selector.flipX();
    else if (source == flipYButton)
      selector.flipY();
    else if (source == rotateClockwiseButton)
      selector.rotateClockwise();
    else if (source == rotateAntiClockwiseButton)
      selector.rotateAntiClockwise();
    else if (source == selectAllButton)
    {
      selector.selectAll();
      setCut(true);
    }
    else if (source == centerButton)
      selector.center();
    else if (source == saveButton)
      writePattern();
    else if (source == selectionRandomButton)
      selector.random();
    else if (source == moveCopyButton)
      if (selector.getCut())
        setCut(false);
      else
        setCut(true);
    requestFocus();
  }

  public void mousePressed(MouseEvent e)
  {
    if (running)
      return;
    if (e.isControlDown())
    {
      if (selector.getState() == Selector.State.None)
        selector.setTopLeft(e.getX(), e.getY());
      else if (selector.getState() == Selector.State.TopLeftMarked)
      {
        selector.save(e.getX(), e.getY());
        setCut(true);
      }
      else if (selector.getState() == Selector.State.GroupSelected)
      {
        selector.restoreCells();
        selector.remove();
      }
    }
    else if (selector.getState() == Selector.State.None)
    {
      if (e.isShiftDown())
      {
        triGrid.changeCurrentState();
        colorSelector.showCurrentState();
      }
      triGrid.setCellState(e.getX(), e.getY());
    }
  }

  public void mouseDragged(MouseEvent e)
  {
    if (selector.getState() == Selector.State.GroupSelected)
    { 
      if (selector.move(e.getX(), e.getY()))
        triGrid.setPosition(e.getX(), e.getY());
    }
  }

  public void mouseMoved(MouseEvent e)
  {
    triGrid.setPosition(e.getX(), e.getY());
  }

  public void mouseExited(MouseEvent e) {}
  public void mouseClicked(MouseEvent e) {}
  public void mouseEntered(MouseEvent e) {}
  public void mouseReleased(MouseEvent e) {}

  public void keyPressed(KeyEvent e)
  {
    int keyCode = e.getKeyCode();
    if (running)
    {
      if (keyCode == KeyEvent.VK_ENTER)
      {
        stopRunner();
        return;
      }
      else if (keyCode == KeyEvent.VK_N)
      {
        stopRunner();
        rule.setNewRule();
        random(columns, rows, cells);
        startRunner();
        return;
      }
      else if (keyCode == KeyEvent.VK_R)
      {
        stopRunner();
        random(columns, rows, cells);
        startRunner();
        return;
      }
      else if (keyCode == KeyEvent.VK_ADD)
      {
        if (delay >= 10)
          delay -= 10;
      }
      else if (keyCode == KeyEvent.VK_SUBTRACT)
        delay += 10;
      statusBar.setSpeedDelay(delay);
      return;
    }
    if (keyCode == KeyEvent.VK_ENTER)
    {
      if (! analyzing)
        startRunner();
      return;
    }
    else if (keyCode == KeyEvent.VK_SPACE)
    {
      step();
      triGrid.repaint();
      return;
    }
    else if (keyCode == KeyEvent.VK_N)
    {
      rule.setNewRule();
      random(columns, rows, cells);
      startRunner();
    }
    else if (keyCode == KeyEvent.VK_R)
    {
      random(columns, rows, cells);
    }
    int increment = (e.isShiftDown()) ? 20 : 2;
    if (keyCode == KeyEvent.VK_LEFT)
      triGrid.decColumnOffset(increment);
    else if (keyCode == KeyEvent.VK_RIGHT)
      triGrid.incColumnOffset(increment);
    else if (keyCode == KeyEvent.VK_UP)
      triGrid.decRowOffset(increment);
    else if (keyCode == KeyEvent.VK_DOWN)
      triGrid.incRowOffset(increment);
    else if (keyCode == KeyEvent.VK_ADD)
      triGrid.zoomIn();
    else if (keyCode == KeyEvent.VK_SUBTRACT)
      triGrid.zoomOut();
    else if (keyCode == KeyEvent.VK_PAGE_UP)
    {
      if ((patterns != null) && (patterns.getNumberPatterns() > 0))
      {
        clear();
        File f = patterns.getPrevious();
        readFile(f);
      }
    }
    else if (keyCode == KeyEvent.VK_PAGE_DOWN)
    {
      if ((patterns != null) && (patterns.getNumberPatterns() > 0))
      {
        clear();
        File f = patterns.getNext();
        readFile(f);
      }
    }
    else if (keyCode == KeyEvent.VK_HOME)
    {
      if ((patterns != null) && (patterns.getNumberPatterns() > 0))
      {
        clear();
        File f = patterns.getCurrent();
        readFile(f);
      }
    }
}

  public void keyReleased(KeyEvent e) {}
  public void keyTyped(KeyEvent e) {}
}